package com.lemoulinstudio.jnag.rds;

import com.lemoulinstudio.jnag.common.BindToLocalId;
import com.lemoulinstudio.jnag.common.BindToSharedId;
import com.lemoulinstudio.jnag.common.Local;
import com.sun.sgs.app.AppContext;
import com.sun.sgs.app.ManagedObjectRemoval;
import com.sun.sgs.app.ManagedReference;
import java.io.ByteArrayOutputStream;
import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.IOException;
import java.io.Serializable;
import java.nio.ByteBuffer;
import java.util.logging.Logger;
import com.lemoulinstudio.jnag.common.MessageListener;
import com.lemoulinstudio.jnag.common.Remote;
import com.lemoulinstudio.jnag.common.Singleton;
import com.lemoulinstudio.jnag.utils.ByteBufferInputStream;
import com.sun.sgs.app.DataManager;
import java.util.HashMap;
import java.util.Map;

/**
 * An implementation of the {@link JnagSession} interface.
 *
 * @author Vincent Cantin
 */
public class JnagSessionImpl implements JnagSession, ManagedObjectRemoval, Serializable {

  private static final long serialVersionUID = 1L;

  public static Logger logger = Logger.getLogger(JnagSession.class.getName());
  private static String sharedIdStoreBindingName = "JnagSharedIdStore";
  private static String configBindingName = "JnagConfig";

  /**
   * Initializes some objects used globally by the Jnag system.
   * 
   * @param configuration An instance of the configuration class
   * generated by the annotation processor.
   */
  public static void initializeGlobalObjects(AbstractConfiguration configuration) {
    AppContext.getDataManager().setBinding(sharedIdStoreBindingName, new SharedIdStore());

    AppContext.getDataManager().setBinding(configBindingName, configuration);

    ProxyFactoryImpl.initializeStaticObjects(configuration);
    SharedIdObjectBinding.initializeStaticObjects();
  }

  private static ManagedReference<LocalIdStore> getSharedIdStoreRef() {
    return AppContext.getDataManager().createReference((LocalIdStore)
            AppContext.getDataManager().getBinding(sharedIdStoreBindingName));
  }

  private static AbstractConfiguration getConfiguration() {
    return (AbstractConfiguration) AppContext.getDataManager().getBinding(configBindingName);
  }

  private Map<String, ManagedReference<? extends Local>> singletonClassNameToLocalSingletonRef;
  private Map<ManagedReference<? extends Local>, String> localSingletonRefToSingletonClassName;

  private Map<String, ManagedReference<? extends Remote>> singletonClassNameToRemoteSingletonRef;
  private Map<ManagedReference<? extends Remote>, String> remoteSingletonRefToSingletonClassName;

  private ManagedReference<LocalIdStore> sharedIdStoreRef;
  private IdObjectBinding<BindToSharedId> sharedIdObjectBinding;
  
  private LocalIdStore localIdStore;
  private IdObjectBinding<BindToLocalId> localIdObjectBinding;
  
  private RemoteIdStore remoteIdStore;
  private IdObjectBinding<Remote> remoteIdObjectBinding;
  
  private ProxyFactory proxyFactory;
  private RootDecoder rootDecoder;
  
  private MessageListener messageListener;
  private ManagedReference<?> callerObjectRef;

  /**
   * Creates a Jnag session.
   */
  public JnagSessionImpl() {
    singletonClassNameToLocalSingletonRef = new HashMap<String, ManagedReference<? extends Local>>();
    localSingletonRefToSingletonClassName = new HashMap<ManagedReference<? extends Local>, String>();

    singletonClassNameToRemoteSingletonRef = new HashMap<String, ManagedReference<? extends Remote>>();
    remoteSingletonRefToSingletonClassName = new HashMap<ManagedReference<? extends Remote>, String>();

    sharedIdStoreRef = getSharedIdStoreRef();
    sharedIdObjectBinding = new SharedIdObjectBinding();
    
    localIdStore = new LocalIdStoreImpl();
    localIdObjectBinding = new IdObjectBindingImpl<BindToLocalId>();

    remoteIdStore = new RemoteIdStoreImpl();
    remoteIdObjectBinding = new IdObjectBindingImpl<Remote>();
    
    proxyFactory = new ProxyFactoryImpl();
    rootDecoder = getConfiguration().getRootDecoder();
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void decodeAndExecute(ByteBuffer binaryMessage) {
    DataInputStream inputStream = new DataInputStream(new ByteBufferInputStream(binaryMessage));

    // Decode the message and invoke the method.
    try {rootDecoder.decodeAndInvoke(this, inputStream);}
    catch (Exception e) {e.printStackTrace();}
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void setMessageListener(MessageListener messageListener) {
    this.messageListener = messageListener;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void setCallerObject(ManagedReference<?> callerObjectRef) {
    this.callerObjectRef = callerObjectRef;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public ManagedReference<?> getCallerObject() {
    return callerObjectRef;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public <T extends Singleton & Local> void bindSingleton(ManagedReference<? extends T> localObjectRef, Class<T> localInterface) {
    assert !singletonClassNameToLocalSingletonRef.containsKey(localInterface.getName()) : "A singleton is already registered for this class.";
    singletonClassNameToLocalSingletonRef.put(localInterface.getName(), localObjectRef);
    localSingletonRefToSingletonClassName.put(localObjectRef, localInterface.getName());
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public <T extends BindToSharedId & Local> void bindToSharedId(ManagedReference<? extends T> localObjectRef, Class<T> localInterface) {
    Id sharedId = sharedIdStoreRef.getForUpdate().createId();
    sharedIdObjectBinding.bind(sharedId, localObjectRef);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public <T extends BindToLocalId & Local> void bindToLocalId(ManagedReference<? extends T> localObjectRef, Class<T> localInterface) {
    Id localId = localIdStore.createId();
    localIdObjectBinding.bind(localId, localObjectRef);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public <T extends Singleton & Local> void unbindSingleton(ManagedReference<? extends T> localObjectRef) {
    String singletonClassName = localSingletonRefToSingletonClassName.remove(localObjectRef);
    assert singletonClassName != null : "This singleton is not bound.";
    singletonClassNameToLocalSingletonRef.remove(singletonClassName);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public <T extends BindToSharedId & Local> void unbindFromSharedId(ManagedReference<? extends T> localObjectRef) {
    Id id = sharedIdObjectBinding.unbindFromObjectRef(localObjectRef);
    sharedIdStoreRef.getForUpdate().releaseId(id);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public <T extends BindToLocalId & Local> void unbindFromLocalId(ManagedReference<? extends T> localObjectRef) {
    Id id = localIdObjectBinding.unbindFromObjectRef(localObjectRef);
    localIdStore.releaseId(id);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public <T extends Singleton & Remote> ManagedReference<T> createSingletonProxy(Class<T> remoteInterface) {
    assert !singletonClassNameToRemoteSingletonRef.containsKey(remoteInterface.getName()) : "There is already a singleton bound to this class.";
    ManagedReference<T> proxyRef = AppContext.getDataManager().createReference(proxyFactory.createProxy(this, remoteInterface));
    remoteSingletonRefToSingletonClassName.put(proxyRef, remoteInterface.getName());
    singletonClassNameToRemoteSingletonRef.put(remoteInterface.getName(), proxyRef);
    return proxyRef;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public <T extends BindToLocalId & Remote> ManagedReference<T> createProxy(Class<T> remoteInterface) {
    return createProxy(remoteInterface, remoteIdStore.allocateNewId());
  }

  private <T extends Remote> ManagedReference<T> createProxy(Class<T> remoteInterface, Id id) {
    ManagedReference<T> proxyRef = AppContext.getDataManager().createReference(proxyFactory.createProxy(this, remoteInterface));
    remoteIdObjectBinding.bind(id, proxyRef);
    return proxyRef;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public <T extends Singleton & Remote> void releaseSingletonProxy(ManagedReference<? extends T> proxyRef) {
    String singletonClassName = remoteSingletonRefToSingletonClassName.remove(proxyRef);
    assert singletonClassName != null : "This proxy is not bound to a singleton class.";
    singletonClassNameToRemoteSingletonRef.remove(singletonClassName);
    AppContext.getDataManager().removeObject(proxyRef.get());
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public <T extends BindToSharedId & Remote> void releaseSharedIdProxy(ManagedReference<? extends T> proxyRef) {
    releaseProxy(proxyRef);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public <T extends BindToLocalId & Remote> void releaseLocalIdProxy(ManagedReference<? extends T> proxyRef) {
    releaseProxy(proxyRef);
  }

  private <T extends Remote> void releaseProxy(ManagedReference<? extends T> proxyRef) {
    Id id = remoteIdObjectBinding.unbindFromObjectRef(proxyRef);
    remoteIdStore.releaseId(id);
    AppContext.getDataManager().removeObject(proxyRef.get());
  }
  
  private void encodeObjectId(Id id, DataOutput output) throws IOException {
    output.writeInt(id.getValue());
  }

  private Id decodeObjectId(DataInput input) throws IOException {
    return new Id(input.readInt());
  }

  /**
   * This should be only called from the generated proxy.
   */
  public <T extends BindToSharedId & Local> void encodeObjectSharedId(ManagedReference<? extends T> objectRef, DataOutput output) throws IOException {
    // Find the associated Id.
    Id id = sharedIdObjectBinding.getId(objectRef);
    assert id != null : "The object is not bound.";

    // Encode it in the output.
    encodeObjectId(id, output);
  }

  /**
   * This should be only called from a generated proxy.
   */
  public <T extends BindToLocalId & Local> void encodeObjectLocalId(ManagedReference<? extends T> objectRef, DataOutput output) throws IOException {
    // Find the associated Id.
    Id id = localIdObjectBinding.getId(objectRef);
    assert id != null : "The object is not bound.";

    // Encode it in the output.
    encodeObjectId(id, output);
  }

  /**
   * This should be only called from a generated proxy.
   */
  public <T extends Remote> void encodeRemoteObjectRef(ManagedReference<? extends T> objectRef, DataOutput output) throws IOException {
    // Find the associated Id.
    Id id = remoteIdObjectBinding.getId(objectRef);
    assert id != null : "The object is not bound.";

    // Encode it in the output.
    encodeObjectId(id, output);
  }

  /**
   * This should be only called from the generated proxy.
   */
  public void sendMessage(ByteArrayOutputStream byteArrayStream) {
    messageListener.notifyMessage(ByteBuffer.wrap(byteArrayStream.toByteArray()));
  }

  /**
   * This should be only called from a generated decoder.
   */
  public <T extends Local & Singleton> ManagedReference<T> getLocalSingletonRef(Class<T> singletonClass) throws IOException {
    assert singletonClassNameToLocalSingletonRef.containsKey(singletonClass.getName()) : "No singleton is bound to this class.";
    return (ManagedReference<T>) singletonClassNameToLocalSingletonRef.get(singletonClass.getName());
  }

  /**
   * This should be only called from a generated decoder.
   */
  public <T extends Remote & Singleton> ManagedReference<T> getRemoteSingletonRef(Class<T> singletonClass) throws IOException {
    ManagedReference<T> singletonRef = (ManagedReference<T>) singletonClassNameToRemoteSingletonRef.get(singletonClass.getName());
    if (singletonRef == null) singletonRef = createSingletonProxy(singletonClass);
    return singletonRef;
  }

  /**
   * This should be only called from a generated decoder.
   */
  public <T extends BindToSharedId & Local> ManagedReference<T> decodeObjectSharedId(DataInput input, Class<T> localClass) throws IOException {
    Id id = decodeObjectId(input);
    assert sharedIdObjectBinding.isUsing(id) : "No object is bound to this id.";
    return (ManagedReference<T>) sharedIdObjectBinding.getObjectRef(id);
  }

  /**
   * This should be only called from a generated decoder.
   */
  public <T extends BindToLocalId & Local> ManagedReference<T> decodeObjectLocalId(DataInput input, Class<T> localClass) throws IOException {
    Id id = decodeObjectId(input);
    assert localIdObjectBinding.isUsing(id) : "No object is bound to this id.";
    return (ManagedReference<T>) localIdObjectBinding.getObjectRef(id);
  }

  /**
   * This should be only called from a generated decoder.
   */
  public <T extends Remote> ManagedReference<T> decodeRemoteObjectRef(DataInput input, Class<T> remoteClass) throws IOException {
    // Decode the Id from the input.
    Id id = decodeObjectId(input);

    // Find the associated object and its class.
    ManagedReference<T> objectRef = (ManagedReference<T>) remoteIdObjectBinding.getObjectRef(id);
    if (objectRef != null) return objectRef;

    // If not found, create a proxy.
    else return createProxy(remoteClass, id);
  }

  /**
   * Logs a text using the default logger of this session.
   *
   * @param text The text to be logged.
   */
  public void logText(String text) {
    logger.info(text);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void removingObject() {
    // Remove the proxies from the data manager.
    remoteIdObjectBinding.removingObject();

    // Remove the singleton proxies from the data manager.
    DataManager dataManager = AppContext.getDataManager();
    for (ManagedReference<? extends Remote> singletonProxyRef : remoteSingletonRefToSingletonClassName.keySet())
      dataManager.removeObject(singletonProxyRef.get());
  }

}
